# pyright: reportArgumentType=false
from typing import cast, Union

from relationalai.early_access.dsl.bindings.tables import (FilteringSubtypeBinding, ConceptBinding, IdentifierBinding,
                                                           SubtypeBinding, RoleBinding, Binding, ColumnRef,
                                                           SnowflakeTable)
from relationalai.early_access.dsl.codegen.common import PotentiallyBoundRelationship, BoundExternalPreferredUC
from relationalai.early_access.dsl.core.types import AbstractValueType
from relationalai.early_access.dsl.core.types import Type
from relationalai.early_access.dsl.ontologies.constraints import Unique, Mandatory
from relationalai.early_access.dsl.ontologies.relationships import Relationship
from relationalai.early_access.dsl.ontologies.roles import Role
from relationalai.early_access.dsl.types.entities import EntityType
from relationalai.rel import OrderedSet


class Reasoner:
    _subtype_map: dict['EntityType', OrderedSet['EntityType']] = {}
    _supertype_map: dict['EntityType', 'EntityType'] = {}
    _subtype_closure: dict[Type, OrderedSet[Type]] = {}
    _supertype_closure: dict[Type, OrderedSet[Type]] = {}
    _type_closure: dict[Type, OrderedSet[Type]] = {}
    # Bindings
    _value_type_bindings: OrderedSet[Binding] = OrderedSet()
    _entity_type_bindings: OrderedSet[Binding] = OrderedSet()
    _constructor_bindings: OrderedSet[Union[IdentifierBinding, SubtypeBinding]] = OrderedSet()
    _filtering_bindings: OrderedSet[FilteringSubtypeBinding] = OrderedSet()
    _referent_constructor_bindings: OrderedSet[Union[IdentifierBinding, SubtypeBinding]] = OrderedSet()
    _referent_bindings: OrderedSet[RoleBinding] = OrderedSet()
    _subtype_bindings: OrderedSet[SubtypeBinding] = OrderedSet()
    _subtype_binding_references: dict[SubtypeBinding, Role] = {}
    _constructor_binding_binds_to: dict[IdentifierBinding, Role] = {}
    _role_bound_thru: dict[Role, list[Binding]] = {}
    _subtype_bound_thru: dict[EntityType, list[Binding]] = {}
    _ref_binding_to_ctor_binding: dict[Binding, Binding] = {}
    _concept_binding_to_actual: dict[Binding, Binding] = {}

    # Relationships & Ontological Constraints
    _unary_relationships: OrderedSet['Relationship'] = OrderedSet()
    _binary_relationships: OrderedSet['Relationship'] = OrderedSet()
    _identifier_relationships: OrderedSet['Relationship'] = OrderedSet()
    _entity_identifier_relationships: OrderedSet['Relationship'] = OrderedSet()
    _entity_type_identifier: dict['EntityType', 'Relationship'] = {}
    _bound_relationships: OrderedSet['PotentiallyBoundRelationship'] = OrderedSet()
    _bound_external_ucs: OrderedSet['BoundExternalPreferredUC'] = OrderedSet()
    _internal_preferred_uc: OrderedSet['Unique'] = OrderedSet()
    _external_preferred_uc: OrderedSet['Unique'] = OrderedSet()
    _constructor_roles: OrderedSet['Role'] = OrderedSet()
    _internal_uc_roles: OrderedSet['Role'] = OrderedSet()
    _mandatory_roles: OrderedSet['Role'] = OrderedSet()
    _bound_role: dict['Role', list['Binding']] = {}
    _ref_type_of: dict['EntityType', 'EntityType'] = {}
    _inclusive_entity_types: OrderedSet['EntityType'] = OrderedSet()
    _composite_entity_types: OrderedSet['EntityType'] = OrderedSet()

    _col_ref_to_pid_concept: dict['ColumnRef', 'EntityType'] = {}

    def __init__(self, model):
        self._model = model

    def value_type_bindings(self):
        return self._value_type_bindings

    def constructor_bindings(self):
        return self._constructor_bindings

    def referent_constructor_bindings(self):
        return self._referent_constructor_bindings

    def subtype_bindings(self):
        return self._subtype_bindings

    def subtype_bindings_of(self, subtype: Type):
        assert isinstance(subtype, EntityType)
        return self._subtype_bound_thru[subtype]

    def bound_relationships(self):
        return self._bound_relationships

    def bound_external_ucs(self):
        return self._bound_external_ucs

    def referent_bindings(self):
        return self._referent_bindings

    def role_bindings(self):
        return self._bound_role

    def role_bindings_of(self, role: Role):
        return self._role_bound_thru[role]

    def get_ref_scheme_type(self, entity_type: Type):
        assert isinstance(entity_type, EntityType)
        return self._ref_type_of.get(entity_type)

    def get_ctor_binding(self, ref_binding: Binding):
        return self._ref_binding_to_ctor_binding[ref_binding]

    def referenced_concept(self, col_ref: ColumnRef):
        return self._col_ref_to_pid_concept.get(col_ref)

    def analyze(self):
        self._analyze_subtypes()
        self._analyze_constraints()
        potentially_bound_rels = self._analyze_bindings()
        self._analyze_ctor_roles()
        self._analyze_relationships()
        self._analyze_bound_external_ucs()
        self._analyze_bound_relationships(potentially_bound_rels)

        self._process_bound_relationships()
        self._process_referent_bindings()

    def _analyze_subtypes(self):
        for subtype_arrow in self._model._subtype_arrows:
            parent = subtype_arrow.end
            if parent not in self._subtype_map:
                self._subtype_map[parent] = OrderedSet()
            self._subtype_map[parent].add(subtype_arrow.start)
            self._supertype_map[subtype_arrow.start] = parent

        self._subtype_closure = {parent: OrderedSet() for parent in self._subtype_map}
        for parent in self._subtype_map.keys():
            self._subtype_closure_dfs(parent, parent)

    def _subtype_closure_dfs(self, prt, curr):
        for chld in self._subtype_map.get(curr, []):
            cls = self._subtype_closure[prt]
            if chld not in cls:
                cls.add(chld)
                self._subtype_closure_dfs(prt, chld)

    def _analyze_bindings(self):
        potentially_bound_rels = {}
        for binding in self._model._bindings:
            role = self._process_binding(binding)
            # if role is None then such binding doesn't bind to a role
            if role is None:
                continue

            if isinstance(binding, ConceptBinding):
                binding = self._concept_binding_to_actual[binding]

            if role not in self._bound_role:
                self._bound_role[role] = []
            self._bound_role[role].append(binding)
            self._categorize_binding(binding, role)

            # Group bindings by relationship and table
            rel = PotentiallyBoundRelationship(role.part_of, binding.column.table)
            if rel not in potentially_bound_rels:
                potentially_bound_rels[rel] = []
            potentially_bound_rels[rel].append(binding)
        return potentially_bound_rels

    def _process_binding(self, binding: 'Binding'):
        """Dispatch binding to the appropriate type-specific handler."""
        if isinstance(binding, ConceptBinding):
            role = self._handle_concept_binding(binding)
        elif isinstance(binding, FilteringSubtypeBinding):
            return self._handle_filtering_binding(binding)
        elif isinstance(binding, SubtypeBinding):
            role = self._handle_subtype_binding(binding)
        elif isinstance(binding, IdentifierBinding):
            role = self._handle_identifier_binding(binding)
        elif isinstance(binding, RoleBinding):
            role = self._handle_role_binding(binding)
        else:
            raise Exception(f'{binding} is yet not supported')

        if role is None:
            raise Exception(f'Unable to lookup binding role for {binding}')
        return role

    def _handle_concept_binding(self, binding: 'ConceptBinding'):
        entity_type = binding.entity_type
        if entity_type in self._inclusive_entity_types:
            actual_binding = IdentifierBinding(binding.column, entity_type=entity_type)
        else:
            actual_binding = SubtypeBinding(binding.column, sub_type=entity_type)
        # Store the actual binding for later reference
        self._concept_binding_to_actual[binding] = actual_binding
        return self._process_binding(actual_binding)

    def _handle_filtering_binding(self, binding: 'FilteringSubtypeBinding'):
        """Process FilteringSubtypeBinding and update filtering and subtype sets."""
        binding = cast(FilteringSubtypeBinding, binding)
        self._filtering_bindings.add(binding)
        self._subtype_bindings.add(binding)
        sub_type = binding.sub_type
        self._ref_type_of[sub_type] = sub_type
        # Initialize list if subtype is encountered for the first time
        if sub_type not in self._subtype_bound_thru:
            self._subtype_bound_thru[sub_type] = []
        self._subtype_bound_thru[sub_type].append(binding)
        return None  # Explicitly return None for filtering bindings

    def _handle_subtype_binding(self, binding: 'SubtypeBinding'):
        """Process SubtypeBinding by looking up ref type and role."""
        ref_type = self._lookup_ref_type_of_subtype(binding.sub_type)
        role = self._lookup_ctor_role(ref_type)
        self._subtype_binding_references[binding] = role
        self._subtype_bindings.add(binding)
        self._ref_type_of[binding.sub_type] = ref_type
        return role

    def _handle_identifier_binding(self, binding: 'IdentifierBinding'):
        """Process IdentifierBinding by looking up its constructor role."""
        role = self._lookup_ctor_role(binding.entity_type)
        self._constructor_binding_binds_to[binding] = role
        return role

    @staticmethod
    def _handle_role_binding(binding: 'RoleBinding'):
        """Process RoleBinding by directly using the role attribute."""
        return binding.role

    def _lookup_ref_type_of_subtype(self, sub_type: 'EntityType'):
        try:
            self._model.identifier_of(sub_type)
            return sub_type
        except KeyError:
            # subtype doesn't have a preferred identifier, so we need to look up the supertype
            supertype = self._supertype_map.get(sub_type)
            if supertype is None:
                raise Exception(f'Subtype {sub_type.name()} has no supertype, cannot infer the reference scheme')
            else:
                return self._lookup_ref_type_of_subtype(supertype)

    def _categorize_binding(self, binding: 'Binding', role: 'Role'):
        player = role.player()
        if isinstance(player, AbstractValueType):
            self._value_type_bindings.add(binding)
        elif isinstance(player, EntityType):
            self._entity_type_bindings.add(binding)
        else:
            raise Exception(f'Binding {binding} is not supported')

    def _analyze_constraints(self):
        for constraint in self._model.constraints():
            if isinstance(constraint, Unique):
                self._process_uc(constraint)
            elif isinstance(constraint, Mandatory):
                self._mandatory_roles.update(constraint.roles())

    def _process_uc(self, constraint: 'Unique'):
        if not constraint.is_preferred_identifier:
            return

        roles = constraint.roles()
        self._constructor_roles.update(roles)  # Mark as constructor roles

        num_roles = len(roles)
        if num_roles == 1:
            self._process_internal_uc(constraint)
        elif num_roles > 1:
            self._external_preferred_uc.add(constraint)

    def _process_internal_uc(self, constraint: Unique):
        self._internal_preferred_uc.add(constraint)

        roles = constraint.roles()
        constructor_role = roles[0]
        rel = constructor_role.part_of

        if rel.arity() != 2:
            raise Exception('Invalid Identifier relationship configuration')

        player = constructor_role.player()
        if isinstance(player, AbstractValueType):
            self._identifier_relationships.add(rel)
        elif isinstance(player, EntityType):
            self._entity_identifier_relationships.add(rel)
        else:
            raise Exception(f'Identifier relationship {rel.pprint()} has unsupported player type {player.name()}')

        concept = constructor_role.sibling().player()
        self._entity_type_identifier[concept] = rel
        self._inclusive_entity_types.add(concept)

    def _analyze_ctor_roles(self):
        for constructor_role in self._constructor_roles:
            self._classify_constructor_bindings(constructor_role)

    def _classify_constructor_bindings(self, constructor_role: 'Role'):
        if constructor_role not in self._bound_role:
            return

        for binding in self._bound_role[constructor_role]:
            if isinstance(binding, IdentifierBinding):
                if not binding.column.references:
                    role = self.lookup_binding_role(binding)
                    self._col_ref_to_pid_concept[binding.column.ref()] = role.player()
                    self._constructor_bindings.add(binding)
                else:
                    self._referent_constructor_bindings.add(binding)
            elif isinstance(binding, SubtypeBinding):
                self._referent_constructor_bindings.add(binding)

    def _analyze_relationships(self):
        for relationship in self._model.relationships():
            arity = relationship.arity()
            if arity == 1:
                self._unary_relationships.add(relationship)
            elif arity == 2:
                self._binary_relationships.add(relationship)

    def _analyze_bound_relationships(self, potentially_bound_rels):
        for rel, bindings in potentially_bound_rels.items():
            rel.bindings.extend(bindings)
            self._analyze_bound_relationship(rel)

    def _lookup_ctor_role(self, entity_type: 'EntityType'):
        # TODO: implement bottom up lookup
        id_rel = self._model.identifier_of(entity_type)
        if id_rel is None:
            raise Exception(f'Identifier relationship for {entity_type} not found')
        for role in id_rel.roles():
            if role.player() == entity_type:
                # ctor role is the only sibling of the constructed entity type role
                return role.sibling()

    def _analyze_bound_relationship(self, meta: 'PotentiallyBoundRelationship'):
        rel = meta.relationship
        arity = rel.arity()

        if arity == 1:
            # Handle unary relationships
            if meta.bindings:
                self._bound_relationships.add(meta)
            return

        if arity != 2:
            raise Exception('N-ary (3+) bound relationship case not supported yet')

        # Handle binary relationships
        key_role, value_role = self._identify_roles(rel)
        key_type, value_type = key_role.player(), value_role.player()
        if not isinstance(key_type, EntityType):
            raise Exception(f'Key role played by {key_type.name()} must be an entity type')

        # TODO : this below should really check if they bound for the same table
        key_role_is_bound = key_role in self._bound_role
        value_role_is_bound = value_role in self._bound_role
        value_role_is_ctor = value_role in self._constructor_roles

        key_role_infers_emap = self._infers_key_role_emap(key_type, meta.table, key_role_is_bound, value_role_is_ctor)
        fully_bound = value_role_is_bound and (key_role_is_bound or key_role_infers_emap)

        is_bound_value_type_relationship = isinstance(value_type, AbstractValueType) and fully_bound
        is_bound = self._is_bound_identifier_relationship(value_role_is_ctor, value_role_is_bound, key_role_is_bound) or \
                   is_bound_value_type_relationship or \
                   self._is_bound_entity_type_relationship(key_type, value_type, key_role, value_role, meta.table)
        if is_bound:
            self._bound_relationships.add(meta)
        else:
            rel_name = meta.relationship.pprint()
            raise Exception(f'Bound relationship `{rel_name}` must have at least one bound role and one inferred entity map')

    def _identify_roles(self, rel: 'Relationship') -> tuple['Role', 'Role']:
        key_role, value_role = None, None
        for role in rel.roles():
            if isinstance(role.player(), AbstractValueType):
                value_role = role
            # if both roles are entity types, arbitrarily pick one as the key role
            elif key_role is None and role not in self._constructor_roles:
                key_role = role
            else:
                value_role = role
        assert key_role is not None
        assert value_role is not None
        return key_role, value_role

    def _infers_key_role_emap(self, key_type: EntityType, table, key_role_is_bound, value_role_is_ctor):
        if key_role_is_bound or value_role_is_ctor:
            return False
        if key_type.is_composite():
            return self._exists_bound_external_uc(key_type, table)
        return self._exists_ctor_binding(key_type, table) or self._exists_ref_binding(key_type, table) or\
                self._exists_subtype_binding(key_type, table)

    def _exists_bound_external_uc(self, key_type, table):
        for bound_constraint in self._bound_external_ucs:
            if bound_constraint.concept == key_type and bound_constraint.table == table:
                return True
        return False

    def _exists_subtype_binding(self, type, table):
        if type not in self._supertype_map:
            return False

        for binding in self._subtype_bindings:
            if binding.column.table == table and binding.sub_type == type:
                ref_type = self._lookup_ref_type_of_subtype(type)
                # TODO: check how this works with multi-table inheritance
                return self._exists_ctor_binding(ref_type, table)
        return False

    @staticmethod
    def _is_bound_identifier_relationship(value_role_is_ctor, value_role_is_bound, key_role_is_bound):
        if value_role_is_ctor and value_role_is_bound:
            if key_role_is_bound:
                raise Exception('Identifier relationship must not have the key role bound')
            return True
        return False

    def _is_bound_entity_type_relationship(self, key_type, value_type, key_role, value_role, table):
        if isinstance(value_type, EntityType) and isinstance(key_type, EntityType):
            value_role_infers_emap = self._exists_ctor_binding(value_type, table) or self._exists_ref_binding(value_type, table)
            return (
                (key_role in self._bound_role and value_role in self._bound_role) or
                (value_role in self._bound_role and value_role_infers_emap) or
                (key_role in self._bound_role and value_role_infers_emap)
            )
        if isinstance(value_type, AbstractValueType) and isinstance(key_type, AbstractValueType):
            raise Exception('Binary Relationship cannot have more than one ValueType role')
        return False

    def _exists_ctor_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        return self.lookup_ctor_binding(entity_type, table) is not None

    def lookup_ctor_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        # TODO : implement more efficient lookup
        for binding, role in self._constructor_binding_binds_to.items():
            sibling_role = role.sibling()
            assert sibling_role is not None
            if sibling_role.player() == entity_type and binding.column.table == table:
                return binding
        if entity_type in self._subtype_closure:
            for subtype in self._subtype_closure[entity_type]:
                cand = self.lookup_ctor_binding(subtype, table)
                if cand:
                    return cand
        return None

    def _exists_ref_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        return self.lookup_ref_binding(entity_type, table) is not None

    def lookup_ref_binding(self, entity_type: 'EntityType', table: 'SnowflakeTable'):
        # TODO : implement more efficient lookup
        for binding, role in self._subtype_binding_references.items():
            if isinstance(binding, SubtypeBinding):
                binding = cast(SubtypeBinding, binding)
                sibling_role = role.sibling()
                assert sibling_role is not None
                ref_type = self._ref_type_of[binding.sub_type]
                if sibling_role.player() == ref_type and binding.column.table == table and entity_type == binding.sub_type:
                    return binding
        if entity_type in self._subtype_closure:
            for subtype in self._subtype_closure[entity_type]:
                cand = self.lookup_ref_binding(subtype, table)
                if cand:
                    return cand
        return None

    def lookup_binding_role(self, binding: 'Binding'):
        if isinstance(binding, IdentifierBinding):
            role = self._constructor_binding_binds_to[binding]
        elif isinstance(binding, RoleBinding):
            role = binding.role
        elif isinstance(binding, FilteringSubtypeBinding):
            ctor_type = self._lookup_ref_type_of_subtype(binding.sub_type)
            role = self._lookup_ctor_role(ctor_type)
        elif isinstance(binding, SubtypeBinding):
            role = self._subtype_binding_references[binding]
        else:
            raise Exception(f'Binding {binding} is not supported')
        if role is None:
            raise Exception(f'Unable to lookup binding role for {binding}')
        return role

    def _process_bound_relationships(self):
        for rel_meta in self._bound_relationships:
            if self.is_identifier_relationship(rel_meta.relationship):
                self._process_identifier_relationship(rel_meta)
            else:
                self._process_non_identifier_relationship(rel_meta)

    def is_identifier_relationship(self, relationship):
        return relationship in self._identifier_relationships or \
               relationship in self._entity_identifier_relationships

    def _process_identifier_relationship(self, rel_meta):
        roles = rel_meta.relationship.roles()
        ctor_role, concept_role = self._classify_identifier_roles(roles)
        ctor_role_bindings, concept_role_bindings = self._categorize_bindings(rel_meta.bindings)
        self._update_role_bound_thru(ctor_role, ctor_role_bindings)
        self._update_role_bound_thru(concept_role, concept_role_bindings)

    def _classify_identifier_roles(self, roles):
        ctor_role = next((role for role in roles if role in self._constructor_roles), None)
        concept_role = next((role for role in roles if role != ctor_role), None)
        return ctor_role, concept_role

    def _categorize_bindings(self, bindings):
        ctor_role_bindings = []
        concept_role_bindings = []
        for binding in bindings:
            if binding in self._constructor_bindings:
                ctor_role_bindings.append(binding)
                concept_role_bindings.append(binding)
            elif binding in self._referent_constructor_bindings:
                ctor_role_bindings.append(binding)
                concept_role_bindings.append(binding)
            elif binding in self._subtype_bindings:
                concept_role_bindings.append(binding)
        return ctor_role_bindings, concept_role_bindings

    def _update_role_bound_thru(self, role, bindings):
        if role not in self._role_bound_thru:
            self._role_bound_thru[role] = []
        self._role_bound_thru[role].extend(bindings)

    def _process_non_identifier_relationship(self, rel_meta):
        role_bindings = {}
        for binding in rel_meta.bindings:
            role = self.lookup_binding_role(binding)
            role_bindings.setdefault(role, []).append(binding)
            if isinstance(binding, RoleBinding) and isinstance(binding.role.player(), EntityType):
                self._referent_bindings.add(binding)
        self._role_bound_thru.update(role_bindings)

    def _process_referent_bindings(self):
        for binding in self._referent_bindings:
            ctor_binding = self._lookup_binding_reference(binding)
            self._ref_binding_to_ctor_binding[binding] = ctor_binding

    def _lookup_binding_reference(self, binding):
        ref_concept = binding.role.player()
        # case 1: ref_concept is an entity type with a reference scheme - just look up the ID relationship
        if ref_concept in self._inclusive_entity_types:
            return self._lookup_inclusive_type_binding(ref_concept, binding)
        # case 2: ref_concept is an entity type with a composite reference scheme
        elif ref_concept in self._composite_entity_types:
            # TODO implement composite reference scheme
            raise Exception('Composite reference scheme not supported yet')
        # case 3: ref_concept has no reference scheme, i.e. exclusive entity type
        else:
            # In this case we look up an IdentifierBinding that has the same column as the binding
            # and role player as the same entity type played by the binding. It must be unique,
            # otherwise we raise an exception.
            #
            # NOTE: this is oversimplified now, as we check subtype closure AND immediate supertype.
            cand_bindings = []
            supertype = self._supertype_map.get(ref_concept)
            if supertype and supertype in self._inclusive_entity_types:
                cand = self._lookup_inclusive_type_binding(supertype, binding)
                if cand:
                    cand_bindings.append(cand)
            else:
                subtype_closure = self._subtype_closure[ref_concept]
                for cand_binding in self._constructor_bindings:
                    ctor_role = self.lookup_binding_role(cand_binding)
                    # to be precise, we need to compare with the sibling role, as it is the one that
                    # is played by the entity type in question
                    sibling_role = ctor_role.sibling()
                    assert sibling_role is not None
                    concept = sibling_role.player()
                    assert isinstance(concept, EntityType)
                    if cand_binding.column == binding.column and concept in subtype_closure:
                        cand_bindings.append(cand_binding)
            if len(cand_bindings) != 1:
                raise Exception(f'Binding {binding} has no unique reference to an IdentifierBinding')
            else:
                return cand_bindings[0]

    def _lookup_inclusive_type_binding(self, ref_concept: 'EntityType', binding: 'Binding'):
        identifier_relationship = self._model.identifier_of(ref_concept)
        # TODO optimize
        ctor_binding = None
        concept_role = None
        for role in identifier_relationship.roles():
            if role in self._constructor_roles:
                concept_role = role.sibling()
            else:
                concept_role = role
            break
        for cand_binding in self._role_bound_thru[concept_role]:
            if isinstance(cand_binding, IdentifierBinding) and binding.column == cand_binding.column:
                ctor_binding = cand_binding
                break
        return ctor_binding

    def _analyze_bound_external_ucs(self):
        for unique_constraint in self._external_preferred_uc:
            roles = unique_constraint.roles()
            role_bindings_by_table = self._collect_table_role_bindings(roles)
            for table, role_bindings in role_bindings_by_table.items():
                if len(role_bindings) < len(roles):
                    raise Exception(
                        f'External preferred identifier {unique_constraint} must have all roles bound'
                    )
                bound_uc = BoundExternalPreferredUC(unique_constraint, table, role_bindings)
                self._bound_external_ucs.add(bound_uc)

    def _collect_table_role_bindings(self, roles):
        role_bindings_by_table = {}
        for role in roles:
            bindings = self._bound_role.get(role)
            if not bindings:
                continue
            for binding in bindings:
                table = binding.column.table
                role_bindings_by_table.setdefault(table, {}).setdefault(role, []).append(binding)
        return role_bindings_by_table



from __future__ import annotations
from dataclasses import dataclass, field
from typing import cast, Optional

from relationalai.early_access.metamodel import ir, factory as f, helpers, visitor, builtins
from relationalai.early_access.metamodel.compiler import Pass, group_tasks
from relationalai.early_access.metamodel.util import OrderedSet, ordered_set


class Flatten(Pass):
    """
    Traverses the model's root to flatten it as much as possible. The result of this pass is
    a Logical root where all nested tasks that represent a rule in Rel are extraced to the
    top level.

    - nested logical with updates becomes a top-level logical (a rule)

    From:
        Logical
            Logical
                lookup1   <- scope is spread
                Logical
                    lookup2
                    derive foo
                Logical
                    lookup3
                    derive bar
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive foo
            Logical
                lookup1
                lookup3
                derive bar

    - nested logical with aggregates becomes a top-level logical (a rule representing an aggregation)

    From:
        Logical
            Logical
                lookup1
                Logical
                    lookup2
                    aggregate1
                Logical
                    lookup3
                    aggregate2
                output
    To:
        Logical
            Logical
                lookup1
                lookup2
                aggregate1
                derive tmp1
            Logical
                lookup1
                lookup3
                aggregate2
                derive tmp2
            Logical
                lookup1
                lookup tmp1
                lookup tmp2
                output

    - a union becomes a top-level logical for each branch, writing into a temporary relation,
    and a lookup from that relation.

    From:
        Logical
            Logical
                Union
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                lookup2
                derive tmp1
            Logical
                lookup tmp1
                output

    - a match becomes a top-level logical for each branch, each writing into its own temporary
    relation and a lookup from the last relation. The top-level logical for a branch derives
    into the temporary relation negating the previous branch:

    From:
        Logical
            Logical
                Match
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                Union            <- tmp1() or (not temp1() and lookup2())
                    lookup tmp1
                    Logical
                        Not
                            lookup tmp1
                        lookup2
                        derive tmp2
            Logical
                lookup tmp2
                output
    """

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, cache) -> ir.Model:

        ctx = Flatten.Context()

        # rewrite the root
        root = self.handle(model.root, ctx)

        # the new body contains the extracted top level logicals and maybe the rewritten root
        body = ctx.top_level if root.replacement is None else ctx.top_level + [root.replacement]

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.relations).frozen(),
            model.types,
            ir.Logical(model.root.engine, tuple(), tuple(body))
        )


    #--------------------------------------------------
    # Helper Classes
    #--------------------------------------------------

    @dataclass
    class ExtractedBindersInfo():
        """ Information about binders extracted from a logical into a top-level logical. """
        common_reference: ir.Lookup
        binders: OrderedSet[ir.Task]
        exposed_vars: OrderedSet[ir.Var]

    class Frame():
        """ Scope information about the logical being analyzed. """
        def __init__(self, parent: Optional[Flatten.Frame], task: ir.Logical):
            # the parent frame, if any
            self.parent = parent
            # binder subtasks
            self.binders: OrderedSet[ir.Task] = ordered_set()
            # grounded subtasks
            self.grounded: OrderedSet[ir.Task] = ordered_set()
            # variables provided by binders or hoisted by composites
            self.locals: OrderedSet[ir.Var] = ordered_set()
            # from task to the locals it provides
            self.task_bindings: dict[ir.Task, OrderedSet[ir.Var]] = dict()
            # from locals to the tasks that provided them
            self.task_by_var: dict[ir.Var, OrderedSet[ir.Task]] = dict()

            # populate the info above
            self._populate(task)

            # may be set by the analysis
            self.extracted_binders: Optional[Flatten.ExtractedBindersInfo] = None


        def intersection(self, tasks: OrderedSet[ir.Task]) -> OrderedSet[ir.Var]:
            """ Return the intersection of variables needed by all task bindings from these tasks. """
            intersection = OrderedSet.from_iterable(self.task_bindings[tasks.some()])
            for t in tasks:
                for v in intersection:
                    if v not in self.task_bindings[t]:
                        intersection.remove(v)
            return intersection

        def local_binders(self, task: ir.Logical) -> list[ir.Task]:
            """ Get all the binders in this frame that do not use hoisted variables from this task. """
            result = []
            hoisted = helpers.hoisted_vars(task.hoisted)
            for binder in self.binders:
                if all([v not in hoisted for v in self.task_bindings[binder]]):
                    result.append(binder)
            return result


        def compute_dependencies(self, vars: OrderedSet[ir.Var], dependencies: OrderedSet[ir.Task]):
            # compute dependencies for these vars
            for var in vars:
                # the var has some outer tasks that depend on it
                if var in self.task_by_var:
                    # if the var is was already in the common extracted logical, no need to bind it again
                    if self.extracted_binders and var in self.extracted_binders.exposed_vars:
                        continue
                    # bring all the binders that this variable needs
                    for t in self.task_by_var[var]:
                        if t in self.binders and t not in dependencies:
                            dependencies.add(t)
                            if t in self.task_bindings:
                                self.compute_dependencies(self.task_bindings[t], dependencies)

        def _populate(self, task):
            """ Analyze this task to populate the frame. """

            # local cache for collect_vars
            collected_vars: dict[ir.Task, OrderedSet[ir.Var]] = dict()

            # populate locals and binders
            for t in task.body:
                if isinstance(t, helpers.BINDERS):
                    self.binders.add(t)
                    collected_vars[t] = helpers.collect_vars(t)
                    self.locals.update(collected_vars[t])
                elif isinstance(t, helpers.COMPOSITES):
                    self.locals.update(helpers.hoisted_vars(t.hoisted))

            # compute all variables in scope
            scope = ordered_set()
            self._collect_variables_in_scope(scope)

            # populate task_bindings
            for t in task.body:
                if isinstance(t, helpers.BINDERS):
                    for v in collected_vars[t]:
                        self._register(self.task_bindings, t, v)
                elif isinstance(t, helpers.COMPOSITES):
                    # - all vars - hoisted - local (not in outer, not in hoisted)
                    expected = (helpers.collect_vars(t) - helpers.hoisted_vars(t.hoisted)) & scope
                    for v in expected:
                        self._register(self.task_bindings, t, v)

            # populate task_by_var
            for t, vars in self.task_bindings.items():
                for v in vars:
                    self._register(self.task_by_var, v, t)

            # populated initially grounded tasks (currently only lookups that are not builtins)
            # TODO: we need a proper dataflow analysis for grounding
            for t in task.body:
                if isinstance(t, ir.Lookup) and not builtins.is_builtin(t.relation):
                    self.grounded.add(t)


        def _collect_variables_in_scope(self, scope: OrderedSet[ir.Var]):
            """ Recursively collect locals from parents. """
            if self.parent:
                self.parent._collect_variables_in_scope(scope)
            scope.update(self.locals)

        def _register(self, map, key, val):
            """ Register key -> val in this map, assuming the map holds ordered sets of vals. """
            if key not in map:
                map[key] = ordered_set()
            map[key].add(val)

    @dataclass
    class Context():
        # the logicals that will be at the top level at the end of the rewrite
        top_level: list[ir.Logical] = field(default_factory=list)
        # new relations created during the pass
        relations: list[ir.Relation] = field(default_factory=list)
        # the stack of frames
        frames: list[Flatten.Frame] = field(default_factory=list)

        def peek(self):
            """ Get the last frame or None. """
            if self.frames:
                return self.frames[-1]
            return None

        def compute_scope(self, task: ir.Task) -> OrderedSet[ir.Task]:
            # get the expected vars for this task on the current frame
            frame = self.peek()
            assert frame
            expected_vars = frame.task_bindings.get(task, ordered_set())

            # compute tasks from previous frames that these vars may depend upon
            tasks = ordered_set()
            for frame in self.frames:
                if frame.extracted_binders:
                    tasks.add(frame.extracted_binders.common_reference)

                if expected_vars:
                    dependencies: OrderedSet[ir.Task] = ordered_set()
                    frame.compute_dependencies(expected_vars, dependencies)
                    tasks.update(dependencies)
            return tasks



    @dataclass
    class HandleResult():
        """ The result of the handle methods. """
        # when a task is handled, the replacement for it, if any; if the task is not changed,
        # this can be the task itself; if the task was extracted as a new logical, for
        # for example, this can be a lookup to the connection relationl; if all sub-tasks
        # became logicals with effects, this can be empty.
        replacement: Optional[ir.Task]

        # True if the task was extracted as its own top-level logical.
        extracted: bool = field(default=False)

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    def handle(self, task: ir.Task, ctx: Context):
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        elif isinstance(task, ir.Union):
            return self.handle_union(task, ctx)
        elif isinstance(task, ir.Match):
            return self.handle_match(task, ctx)
        else:
            return Flatten.HandleResult(task)

    def handle_logical(self, task: ir.Logical, ctx: Context):

        # create frame to process the children
        frame = Flatten.Frame(ctx.peek(), task)
        ctx.frames.append(frame)

        # process the original body
        groups = group_tasks(task.body, {
            "binders": helpers.BINDERS,
            "composites": helpers.COMPOSITES,
            "effects": helpers.EFFECTS
        })

        # potentially extract common binders and register a reference to it in the current frame
        # note that at this point we are only extracting this common relation if the composites
        # are extractable. But once we implement proper handling of nulls, we may need to extract
        # also if the nested logical is nullable.
        common_reference = None
        if len(groups["binders"]) > 1 and len(groups["composites"]) > 1 and len(self._extractables(groups["composites"])) > 1:
            body: OrderedSet[ir.Task] = ordered_set()
            frame.compute_dependencies(helpers.collect_vars(*(frame.grounded)), body)
            exposed_vars = frame.intersection(groups["composites"])
            if len(body) > 1 and exposed_vars:
                common_connection = self._extract(task, body, exposed_vars.list, ctx, "common")
                common_reference = f.lookup(common_connection, exposed_vars.list)
                frame.extracted_binders = Flatten.ExtractedBindersInfo(common_reference, body, exposed_vars)

        # recursively handle children, collecting the replacements in the body
        body:OrderedSet[ir.Task] = ordered_set()
        all_composites_extracted = None
        for t in task.body:
            if t in groups["binders"]:
                body.add(t)
            elif t in groups["composites"]:
                x = self.handle(t, ctx)
                if x.replacement is not None:
                    self._extend_body(body, x.replacement)
                all_composites_extracted = x.extracted and (all_composites_extracted is None or all_composites_extracted)
            else:
                body.add(self.handle(t, ctx).replacement)
                if t not in groups["effects"]:
                    all_composites_extracted = False

        # pop the frame used to process the children
        ctx.frames.pop()

        # if the binders where extracted or pushed down when all other nodes are composites, remove them from the body
        if common_reference or (all_composites_extracted and not groups["effects"] and not groups["other"]):
            body = body - frame.local_binders(task)
        # if the binders where extracted but not pushed down, add a reference to the connection
        if common_reference and not all_composites_extracted:
            body.add(common_reference)

        # all children were extracted, drop it
        if not body:
            return Flatten.HandleResult(None)

        # now process the rewritten body
        groups = group_tasks(body.list, {
            "outputs": ir.Output,
            "updates": ir.Update,
            "aggregates": ir.Aggregate,
        })

        # if there are outputs, currently assume it's already at top level, so just return
        # the rewritten body
        if groups["outputs"]:
            return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body)))

        # if there are updates, extract as a new top level rule
        if groups["updates"]:
            # add whatever was in scope at the start of the body
            body = ctx.compute_scope(task) | body
            ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(body)))

            # no need to refer to the extracte logical because it is an update
            return Flatten.HandleResult(None, True)

        if groups["aggregates"]:
            if len(groups["aggregates"]) > 1:
                # stop rewritting as we don't know how to handle this yet
                return Flatten.HandleResult(task)

            # there must be only one
            agg = cast(ir.Aggregate, groups["aggregates"].some())

            # add whatever was in scope at the start of the body
            scoped = ctx.compute_scope(task)
            body = scoped | body

            # extract a new logical for the aggregate, exposing aggregate group-by and results
            frame = ctx.peek()
            exposed_vars = ordered_set()
            exposed_vars.update(list(agg.group) + helpers.aggregate_outputs(agg))
            # make sure aggregation projection variables are not exposed
            for v in agg.projection:
                exposed_vars.remove(v)
            connection = self._extract(agg, body, exposed_vars.list, ctx)

            # return a reference to the connection relation
            reference = f.logical([f.lookup(connection, exposed_vars.list)], self._merge_var_list(exposed_vars.list, task.hoisted))
            return Flatten.HandleResult(reference, True)

        return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body)))


    def handle_match(self, match: ir.Match, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not match.tasks:
            return Flatten.HandleResult(match)

        body = ctx.compute_scope(match)
        exposed_vars = helpers.collect_vars(*body)
        exposed_vars.update(helpers.hoisted_vars(match.hoisted))
        exposed_vars = exposed_vars.list

        connection = None
        reference = None

        for branch in match.tasks:
            # process the branch
            x = self.handle(branch, ctx)
            assert(x.replacement)

            branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
            self._extend_body(branch_body, x.replacement)
            if reference:
                branch_body.add(self._negate(reference, len(match.hoisted)))
                branch_body = OrderedSet.from_iterable([f.union([f.logical(branch_body.list, match.hoisted), reference], match.hoisted)])
            connection = self._extract(branch, branch_body, exposed_vars, ctx, "match")
            reference = f.logical([f.lookup(connection, exposed_vars)], exposed_vars)

        return Flatten.HandleResult(reference, True)


    def handle_union(self, union: ir.Union, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not union.tasks:
            return Flatten.HandleResult(union)

        body = ctx.compute_scope(union)
        exposed_vars = helpers.collect_vars(*body)
        exposed_vars.update(helpers.hoisted_vars(union.hoisted))
        exposed_vars = exposed_vars.list

        connection = None

        for branch in union.tasks:
            # process the branch
            x = self.handle(branch, ctx)
            assert(x.replacement)

            branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
            self._extend_body(branch_body, x.replacement)

            if connection is None:
                # first branch, extract making a connection relation
                connection = self._extract(branch, branch_body, exposed_vars, ctx, "union")
            else:
                # subsequent branch, extract reusing the connection relation
                # add derivation to the extracted body
                branch_body.add(f.derive(connection, exposed_vars))

                # extract the body
                ctx.top_level.append(ir.Logical(union.engine, tuple(), tuple(branch_body)))

        # return a reference to the connection
        assert(connection)
        reference = f.logical([f.lookup(connection, exposed_vars)], exposed_vars)
        return Flatten.HandleResult(reference, True)

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def _negate(self, reference: ir.Logical, values: int):
        """
        Return a negation of this reference, where the last `values` arguments are to
        be replaced by wildcards (i.e. len(reference.args) - values are keys so they need
        to be bound in the Not.)
        """
        lookup = cast(ir.Lookup, reference.body[0])
        args = []
        i = 0
        last = len(lookup.args) - values
        for arg in lookup.args:
            args.append(f.wild()) if i >= last else args.append(arg)
            i += 1

        return ir.Not(reference.engine, f.lookup(lookup.relation, args))

    def _extractable(self, t: ir.Task):
        """
        Whether this task is a Logical that will be extracted as a top level by this
        pass, because it has an aggregation, effects, match, union, etc.
        """
        return isinstance(t, ir.Logical) and len(visitor.collect_by_type((ir.Update, ir.Aggregate, ir.Match, ir.Union), t)) > 0

    def _extractables(self, composites: OrderedSet[ir.Task]):
        """ Filter the set of composites, keeping only the extractable ones. """
        return list(filter(self._extractable,composites))

    def _extract(self, task: ir.Task, body: OrderedSet[ir.Task], exposed_vars: list[ir.Var], ctx: Context, prefix: Optional[str]=None) -> ir.Relation:
        """
        Extract into this context a new top level Logical that contains this body plus a
        derive task into a new temporary relation, which is also registered with the ctx.
        The exposed_vars determine the arguments of this temporary relation. The prefix
        can be used to customize the name of the relation, which defaults to the task kind.

        Return the temporary relation created for the extraction.
        """

        # TODO: review variable rewrites, i.e. when we extract a new logical, we should remap variables
        p = prefix if prefix else task.kind

        # new relation to derive the aggregation into
        connection = f.relation(f"_{p}_{task.id}", [f.field(v.name, v.type) for v in exposed_vars])
        ctx.relations.append(connection)

        # add derivation to the extracted body
        body.add(f.derive(connection, exposed_vars))

        # extract the body
        ctx.top_level.append(ir.Logical(task.engine, tuple(), tuple(body)))

        # return a reference to the connection relation
        return connection


    def _merge_var_list(self, vars: list[ir.Var], hoisted: tuple[ir.VarOrDefault, ...]) -> list[ir.VarOrDefault]:
        """ Merge vars and hoisted, making sure that hoisted vars have precedence since they may have defaults. """
        r = []
        hoisted_vars = helpers.hoisted_vars(hoisted)
        for v in vars:
            if v not in hoisted_vars:
                r.append(v)
        r.extend(hoisted)
        return r

    def _extend_body(self, body: OrderedSet[ir.Task], extra: ir.Task):
        """ Add the extra task to the body, but if the extra is a simple logical, just
        inline its subtasks. """
        if isinstance(extra, ir.Logical):
            if extra.hoisted:
                # hoists, remove things that are already in the body to avoid duplicates
                logical_body = []
                for t in extra.body:
                    if t not in body:
                        logical_body.append(t)
                if len(logical_body) == len(extra.body):
                    # no duplicates
                    body.add(extra)
                else:
                    # some duplicate, remove them
                    body.add(ir.Logical(
                        extra.engine,
                        extra.hoisted,
                        tuple(logical_body)
                    ))
            else:
                # no hoists, just inline
                body.update(extra.body)
        else:
            body.add(extra)



#--------------------------------------------------
# Compiler
#--------------------------------------------------

from typing import cast, Tuple, Sequence

from relationalai.early_access.metamodel import ir, compiler as c, builtins as rel_builtins
from relationalai.early_access.metamodel import types
from relationalai.early_access.lqp import ir as lqp, llqp
from relationalai.early_access.metamodel.rewrite import Splinter
from relationalai.early_access.metamodel.util import NameCache
from relationalai.early_access.metamodel.compiler import Pass

import hashlib

# TODO: take rewrite out of rel it doesnt belong there
from relationalai.early_access.rel import rewrite

from dataclasses import field

# TODO: assert theres no nested nested nested aggregates
# TODO: assert theres only one aggregate per rule
# TODO: is this correct?
# Move aggregates to the same level as updates (i.e. top-level logical)
class HoistAggregates(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        root = cast(ir.Logical, model.root)
        new_subtasks = []
        for subtask in root.body:
            subtask = cast(ir.Logical, subtask)
            new_subsubtasks = []
            for subsubtask in subtask.body:
                if not isinstance(subsubtask, ir.Logical):
                    new_subsubtasks.append(subsubtask)
                    continue

                subsubtask = cast(ir.Logical, subsubtask)
                if self.has_aggregate(subsubtask):
                    new_subsubtask, aggr = self._hoist_aggregates(subsubtask)
                    new_subsubtasks.append(new_subsubtask)
                    new_subsubtasks.append(aggr)
                else:
                    new_subsubtasks.append(subsubtask)

            new_subtask = ir.Logical(
                    subtask.engine,
                    subtask.hoisted,
                    tuple(new_subsubtasks),
                )
            new_subtasks.append(new_subtask)

        new_root = ir.Logical(root.engine, root.hoisted, tuple(new_subtasks))
        model = ir.Model(
            model.engines,
            model.relations,
            model.types,
            new_root,
        )
        return model

    def has_aggregate(self, task: ir.Task) -> bool:
        assert isinstance(task, ir.Logical)
        for subtask in task.body:
            if isinstance(subtask, ir.Aggregate):
                return True
        return False

    def _hoist_aggregates(self, task: ir.Task) -> Tuple[ir.Task, ir.Aggregate]:
        assert isinstance(task, ir.Logical)
        aggr = None
        new_subtasks = []
        for subtask in task.body:
            if isinstance(subtask, ir.Aggregate):
                assert aggr is None, "only one aggregate supported at the moment"
                aggr = subtask
            else:
                new_subtasks.append(subtask)
        assert aggr is not None, "should have found an aggregate"
        return (ir.Logical(
            task.engine,
            task.hoisted,
            tuple(new_subtasks),
        ), aggr)

# Rewrites aggregates when necessary e.g. avg -> sum/count
class NormalizeAggregates(Pass):
    def rewrite(self, model: ir.Model, cache) -> ir.Model:
        self.var_name_cache: NameCache = NameCache()
        root = cast(ir.Logical, model.root)
        new_subtasks = []
        for subtask in root.body:
            subtask = cast(ir.Logical, subtask)
            if self.has_aggregate(subtask):
                new_subtask = self._rewrite_subtask(subtask)
                new_subtasks.append(new_subtask)
            else:
                new_subtasks.append(subtask)
        new_root = ir.Logical(root.engine, root.hoisted, tuple(new_subtasks))
        model = ir.Model(
            model.engines,
            model.relations,
            model.types,
            new_root,
        )
        return model

    def _rewrite_subtask(self, task: ir.Task) -> ir.Task:
        assert isinstance(task, ir.Logical)
        new_subtasks = []
        for subtask in task.body:
            if isinstance(subtask, ir.Aggregate):
                assert isinstance(subtask, ir.Aggregate)
                if subtask.aggregation.name == "sum":
                    new_subtasks.append(subtask)
                elif subtask.aggregation.name == "count":
                    new_subtasks.append(subtask)
                elif subtask.aggregation.name == "avg":
                    tasks = self._make_avg(subtask)
                    for tt in tasks:
                        new_subtasks.append(tt)
                else:
                    raise NotImplementedError(f"Unsupported aggregate: {subtask.aggregation.name}")
            else:
                new_subtasks.append(subtask)

        return ir.Logical(
            task.engine,
            task.hoisted,
            tuple(new_subtasks),
        )

    def _make_avg(self, task: ir.Aggregate) -> list[ir.Task]:
        assert isinstance(task, ir.Aggregate)
        assert task.aggregation.name == "avg"

        # TODO: confirm input args
        assert len(task.args) == 2
        input_arg = task.args[0]
        output_arg = task.args[1]
        assert isinstance(input_arg, ir.Var)
        assert isinstance(output_arg, ir.Var)

        new_sum_name = self.var_name_cache.get_name(input_arg.id, "sum")
        new_count_name = self.var_name_cache.get_name(input_arg.id, "count")
        new_sum = ir.Var(output_arg.type, new_sum_name)
        new_count = ir.Var(output_arg.type, new_count_name)

        sum_args = tuple([input_arg, new_sum])
        count_args = tuple([input_arg, new_count])

        sum_task = ir.Aggregate(
            task.engine,
            rel_builtins.sum,
            task.projection,
            task.group,
            sum_args,
        )
        count_task = ir.Aggregate(
            task.engine,
            rel_builtins.count,
            task.projection,
            task.group,
            count_args,
        )

        final_sum = ir.Lookup(
            task.engine,
            rel_builtins.plus,
            tuple([new_sum, new_count, output_arg]),
        )

        return [
            sum_task,
            count_task,
            final_sum,
        ]

    def has_aggregate(self, task: ir.Task) -> bool:
        assert isinstance(task, ir.Logical)
        for subtask in task.body:
            if isinstance(subtask, ir.Aggregate):
                return True
        return False

class Compiler(c.Compiler):
    def __init__(self):
        # TODO: are these all needed from the Rel emitter?
        # TODO: should there be a pass to remove aliases from output?
        super().__init__([
            rewrite.Flatten(),

            # Adds missing existentials + splits multi-headed rules into single rules
            Splinter(),

            HoistAggregates(),
            NormalizeAggregates(),
        ])

    # TODO: what should return value be? string according to parents?
    def do_compile(self, model: ir.Model, options:dict={}) -> str:
        lqp_ir, debugging_ctx = Model2Lqp().to_lqp(model)
        self._lqp_result = lqp_ir
        # Compare to False specifically as we want to default (None, not in dict) to styled.
        styled = True if options.get("styled") in [None, True] else False
        lqp_str = llqp.program_to_llqp(lqp_ir, styled)

        debug_str = ""
        if len(debugging_ctx.id_to_orig_name) > 0:
            debug_str += ";; Original names\n"

        for (rid, name) in debugging_ctx.id_to_orig_name.items():
            debug_str += ";; \t " + str(rid) + " -> `" + name + "`\n"

        if debug_str != "":
            lqp_str += "\n\n"
            lqp_str += ";; Debug information\n"
            lqp_str += ";; -----------------------\n"
            lqp_str += debug_str

        return lqp_str

    def get_lqp_ir(self) -> lqp.LqpProgram:
        """ Returns the LQP IR generated by the compiler. """
        assert self._lqp_result is not None, "LQP IR not generated yet"
        return self._lqp_result

class DebuggingCtx:
    """ Extra information only used for debugging. """
    def __init__(self, id_to_orig_name: dict[lqp.RelationId, str]):
       self.id_to_orig_name = id_to_orig_name

class Model2Lqp:
    """ Generates LQP IR from the model IR. """
    def __init__(self):
        # TODO: comment htese fields
        # TODO: should we have a pass to rename variables instead of this?
        self.var_name_cache: NameCache = field(default_factory=NameCache)
        self.id_to_orig_name: dict[lqp.RelationId, str] = field(default_factory=dict)
        self.output_ids: list[lqp.RelationId] = field(default_factory=list)

    """ Main access point. Converts the model IR to an LQP program. """
    def to_lqp(self, model: ir.Model) -> Tuple[lqp.LqpProgram, DebuggingCtx]:
        _assert_valid_input(model)
        self._reset_state()
        program = self._translate_to_program(model)
        debugging_ctx = DebuggingCtx(self.id_to_orig_name)
        return (program, debugging_ctx)

    def _reset_state(self):
        self.var_name_cache = NameCache()
        self.id_to_orig_name = {}
        self.output_ids = []

    def _translate_to_program(self, model: ir.Model) -> lqp.LqpProgram:
        decls: list[lqp.Declaration] = []
        outputs: list[Tuple[str, lqp.RelationId]] = []

        # LQP only accepts logical tasks
        # These are asserted at init time
        root = cast(ir.Logical, model.root)

        for subtask in root.body:
            # TODO: when do we get more than one?
            child_l = cast(ir.Logical, subtask)
            decl = self._translate_to_decl(child_l)
            for d in decl:
                assert isinstance(d, lqp.Declaration)
                decls.append(d)

        for output_id in self.output_ids:
            assert isinstance(output_id, lqp.RelationId)
            outputs.append(("output", output_id))

        return lqp.LqpProgram(decls, outputs)

    def _translate_to_decl(self, rule: ir.Logical) -> list[lqp.Declaration]:
        outputs = []
        updates = []
        body_tasks = []
        aggregates = []

        for task in rule.body:
            if isinstance(task, ir.Output):
                outputs.append(task)
            elif isinstance(task, ir.Lookup):
                body_tasks.append(task)
            elif isinstance(task, ir.Logical):
                body_tasks.append(task)
            elif isinstance(task, ir.Exists):
                body_tasks.append(task)
            elif isinstance(task, ir.Aggregate):
                aggregates.append(task)
            elif isinstance(task, ir.Update):
                updates.append(task)
            else:
                raise NotImplementedError(f"Unknown task type: {type(task)}")

        conjuncts = []
        for task in body_tasks:
            conjunct = self._translate_to_formula(task)
            conjuncts.append(conjunct)
        body = lqp.Conjunction(conjuncts) if len(conjuncts) != 1 else conjuncts[0]

        # TODO: unify all the def/rel-id generation
        if len(aggregates) > 0:
            assert len(outputs) == 0, "cannot have both outputs and aggregates"
            assert len(updates) == 1, "aggregates should derive into another relation"

            update = updates[0]
            assert isinstance(update, ir.Update)
            assert update.effect == ir.Effect.derive, "only derive supported at the moment"

            final_defs = []
            for aggr in aggregates:
                assert isinstance(aggr, ir.Aggregate)
                final_defs.append(self._translate_aggregate(aggr, body, update))

            return final_defs

        assert len(aggregates) == 0, "aggregates should have already been done"

        if len(outputs) > 0:
            assert len(outputs) == 1, "only one output supported at the moment"
            output = outputs[0]
            assert isinstance(output, ir.Output)
            output_vars = []
            for _, v in output.aliases:
                # TODO: we dont yet handle aliases
                assert isinstance(v, ir.Var)
                var = self._translate_to_var(v)
                output_vars.append(var)
            abstraction = lqp.Abstraction(output_vars, body)
            # TODO: is this correct? might need attrs tooo?
            rel_id = _gen_rel_id(abstraction)
            self.output_ids.append(rel_id)
            return [lqp.Def(rel_id, abstraction, [])]

        assert len(updates) == 1, "only one update supported at the moment"
        update = updates[0]
        assert isinstance(update, ir.Update)
        effect = update.effect
        assert effect == ir.Effect.derive, "only derive supported at the moment"

        args = []
        for var in update.args:
            assert isinstance(var, ir.Var)
            args.append(self._translate_to_var(var))

        abstraction = lqp.Abstraction(args, body)
        # TODO: is this correct? might need attrs tooo?
        rel_id = _gen_rel_id(abstraction)
        self.id_to_orig_name[rel_id] = update.relation.name
        return [lqp.Def(rel_id, abstraction, [])]

    def _translate_aggregate(self, aggr: ir.Aggregate, body: lqp.Formula, update: ir.Update) -> lqp.Def:
        assert isinstance(aggr, ir.Aggregate)

        # TODO: hadnle this properly
        aggr_name = aggr.aggregation.name
        assert aggr_name == "sum" or aggr_name == "count", f"only support sum or count for now, not {aggr.aggregation.name}"

        group_bys = []
        for arg in aggr.group:
            assert isinstance(arg, ir.Var)
            group_bys.append(self._translate_to_var(arg))

        projected_args = []
        for arg in aggr.projection:
            assert isinstance(arg, ir.Var)
            projected_args.append(self._translate_to_var(arg))

        # TODO: differentiate between results and more args
        result_args = []
        if aggr_name == "count":
            # TODO: this is wrong?? why sometimes it has two args and sometimes one :(
            assert len(aggr.args) >= 1
            result_args.append(lqp.Constant(1))
            output_arg = aggr.args[0]
            assert isinstance(output_arg, ir.Var)
            result_args.append(self._translate_to_var(output_arg))

        op = self._lqp_operator(aggr.aggregation)
        inner_abstr = lqp.Abstraction(
            group_bys,
            body,
        )
        reduce = lqp.Reduce(
            op,
            inner_abstr,
            result_args,
        )
        abstraction = lqp.Abstraction(
            projected_args,
            reduce,
        )
        rel_id = _gen_rel_id(abstraction)
        self.id_to_orig_name[rel_id] = update.relation.name
        return lqp.Def(rel_id, abstraction, [])

    def _lqp_sum_op(self) -> lqp.Abstraction:
        # TODO: make sure gensym'd properly
        vs = [
            lqp.Var("x", lqp.PrimitiveType.INT),
            lqp.Var("y", lqp.PrimitiveType.INT),
            lqp.Var("z", lqp.PrimitiveType.INT),
        ]

        body = lqp.Primitive("rel_primitive_add", [vs[0], vs[1], vs[2]])

        return lqp.Abstraction(vs, body)

    def _lqp_operator(self, op: ir.Relation) -> lqp.Abstraction:
        if op.name == "sum":
            return self._lqp_sum_op()
        elif op.name == "count":
            return self._lqp_sum_op()
        else:
            raise NotImplementedError(f"Unsupported aggregation: {op.name}")

    def _translate_to_formula(self, task: ir.Task) -> lqp.Formula:
        if isinstance(task, ir.Logical):
            conjuncts = []
            for child in task.body:
                conjunct = self._translate_to_formula(child)
                conjuncts.append(conjunct)
            return lqp.Conjunction(conjuncts)
        elif isinstance(task, ir.Lookup):
            return self._translate_to_atom(task)
        elif isinstance(task, ir.Exists):
            lqp_vars = []
            for var in task.vars:
                lqp_vars.append(self._translate_to_var(var))
            formula = self._translate_to_formula(task.task)
            return lqp.Exists(lqp_vars, formula)
        else:
            raise NotImplementedError(f"Unknown task type (formula): {type(task)}")

    def _translate_to_var(self, var: ir.Var) -> lqp.Var:
        name = self.var_name_cache.get_name(var.id, var.name)
        t = type_from_var(var)
        return lqp.Var(name, t)

    def _translate_to_atom(self, task: ir.Lookup) -> lqp.Formula:
        # TODO: want signature not name
        rel_name = task.relation.name
        assert isinstance(rel_name, str)
        terms = []
        sig_types = []
        for arg in task.args:
            if isinstance(arg, lqp.PrimitiveValue):
                term = lqp.Constant(arg)
                terms.append(term)
                t = type_from_constant(arg)
                sig_types.append(t)
                continue
            assert isinstance(arg, ir.Var)
            var = self._translate_to_var(arg)
            terms.append(var)
            sig_types.append(var.type)

        # TODO: wrong
        if rel_builtins.is_builtin(task.relation):
            return self._translate_builtin_to_primitive(task.relation, terms)

        return lqp.RelAtom(lqp.RelationSig(rel_name, sig_types), terms)

    def _translate_builtin_to_primitive(self, relation: ir.Relation, terms: list[lqp.Term]) -> lqp.Primitive:
        lqp_name = self._name_to_lqp_name(relation.name)
        return lqp.Primitive(lqp_name, terms)

    def _name_to_lqp_name(self, name: str) -> str:
        # TODO: do these proprly
        if name == "+":
            return "rel_primitive_add"
        elif name == "-":
            return "rel_primitive_subtract"
        elif name == "*":
            return "rel_primitive_multiply"
        elif name == "=":
            return "rel_primitive_eq"
        elif name == "<=":
            return "rel_primitive_lt_eq"
        else:
            raise NotImplementedError(f"missing primitive case: {name}")

def _gen_rel_id(abstr: lqp.Abstraction) -> lqp.RelationId:
    return lqp.RelationId(hash_to_uint128(_lqp_hash(abstr)))

# TODO: this is NOT a good hash its just to get things working for now to get
# a stable id.
def _lqp_hash(node: lqp.LqpNode) -> int:
    if isinstance(node, lqp.Abstraction):
        h1 = _lqp_hash_list(node.vars)
        h2 = _lqp_hash(node.value)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.Exists):
        h1 = _lqp_hash_list(node.vars)
        h2 = _lqp_hash(node.value)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.Conjunction):
        h1 = _lqp_hash_list(node.args)
        return _lqp_hash_fn((h1,))
    elif isinstance(node, lqp.Var):
        return _lqp_hash_fn((node.name, node.type))
    elif isinstance(node, lqp.Constant):
        return _lqp_hash_fn((node.value,))
    elif isinstance(node, lqp.RelAtom):
        h1 = _lqp_hash(node.sig)
        h2 = _lqp_hash_list(node.terms)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.RelationSig):
        return _lqp_hash_fn((node.name, tuple(node.types)))
    elif isinstance(node, lqp.Primitive):
        h1 = _lqp_hash_fn(node.name)
        h2 = _lqp_hash_list(node.terms)
        return _lqp_hash_fn((h1, h2))
    elif isinstance(node, lqp.Reduce):
        h1 = _lqp_hash(node.op)
        h2 = _lqp_hash(node.body)
        h3 = _lqp_hash_list(node.terms)
        return _lqp_hash_fn((h1, h2, h3))
    else:
        raise NotImplementedError(f"Unsupported LQP node type: {type(node)}")

# TODO: this is NOT a good hash its just to get things working for now to get
# a stable id.
def _lqp_hash_fn(node) -> int:
    return int.from_bytes(hashlib.sha256(str(node).encode()).digest(), byteorder='big', signed=False)

def _lqp_hash_list(node: Sequence[lqp.LqpNode]) -> int:
    hashes = [_lqp_hash(n) for n in node]
    return hash(tuple(hashes))

def hash_to_uint128(h: int) -> int:
    return h % (2**128)  # Ensure it's within the 128-bit range

# Preconditions
def _assert_valid_input(model: ir.Model) -> bool:
    # TODO: flesh this out more?
    _assert_root_is_logical(model.root)
    return True

def _assert_root_is_logical(task: ir.Task) -> bool:
    assert isinstance(task, ir.Logical), f"expected logical task, got {type(task)}"
    for subtask in task.body:
        # TODO: assert what subtasks should look like
        assert isinstance(subtask, ir.Logical), f"expected logical task, got {type(subtask)}"

    return True

def type_from_var(arg: ir.Var) -> lqp.PrimitiveType:
    assert isinstance(arg, ir.Var)
    assert isinstance(arg.type, ir.ScalarType)
    if types.is_builtin(arg.type):
        # TODO: just ocompare to types.py
        if arg.type.name == "Int":
            return lqp.PrimitiveType.INT
        else:
            # TODO: fix this
            assert arg.type.name == "Any" or arg.type.name == "Number", f"Unknown type: {arg.type.name}"
            return lqp.PrimitiveType.UNKNOWN
    else:
        return lqp.PrimitiveType.UNKNOWN

def type_from_constant(arg: lqp.PrimitiveValue) -> lqp.PrimitiveType:
    if isinstance(arg, int):
        return lqp.PrimitiveType.INT
    elif isinstance(arg, float):
        return lqp.PrimitiveType.FLOAT
    elif isinstance(arg, str):
        return lqp.PrimitiveType.STRING
    else:
        raise NotImplementedError(f"Unknown constant type: {type(arg)}")



"""
Helpers to analyze the metamodel IR.
"""
from __future__ import annotations

from typing import cast, Tuple
from relationalai.early_access.metamodel import ir, util, visitor, builtins
from relationalai.early_access.metamodel.util import OrderedSet, FrozenOrderedSet


#--------------------------------------------------
# Checks
#--------------------------------------------------

def is_external(relation: ir.Relation):
   """ Whether this relation is external, by being marked with the external annotation. """
   return builtins.external_annotation in relation.annotations

def is_aggregate_input(var: ir.Var, agg: ir.Aggregate):
    """ Whether this var is an input to this aggregation. """
    # field at the var index is input (the +2 is for projection + group)
    return (var in agg.args and agg.aggregation.fields[agg.args.index(var) + 2].input)

def is_effective_logical(n: ir.Task):
    """ Whether this task is a Logical and contains an Update child, recursively. """
    return isinstance(n, ir.Logical) and len(visitor.collect_by_type(ir.Update, n)) > 0

def is_nullable_logical(n: ir.Task):
    """ Whether this task is a Logical that contains a hoisted variable with a None default. """
    return isinstance(n, ir.Logical) and any(isinstance(v, ir.Default) and v.value is None for v in n.hoisted)

#--------------------------------------------------
# Filters
#--------------------------------------------------

def aggregate_outputs(agg: ir.Aggregate) -> list[ir.Var]:
    """ Get the list of vars bound to the outputs of this aggregation. """
    return list(filter(lambda arg: isinstance(arg, ir.Var) and not is_aggregate_input(arg, agg), agg.args)) # type: ignore

def effective_logicals(tasks: OrderedSet[ir.Task]) -> OrderedSet[ir.Logical]:
    """ Filter tasks to return only the Logical tasks that are effective. """
    return util.OrderedSet.from_iterable(filter(lambda t: is_effective_logical(t), tasks))

def nullable_logicals(tasks: OrderedSet[ir.Task]) -> OrderedSet[ir.Logical]:
    """ Filter tasks to return only the Logical tasks that are nullable. """
    return util.OrderedSet.from_iterable(filter(lambda t: is_nullable_logical(t), tasks))

def hoisted_vars(hoisted: Tuple[ir.VarOrDefault, ...]) -> list[ir.Var]:
    """ Extract the vars from defaults in the hoisted list, returning just Vars. """
    return [v.var if isinstance(v, ir.Default) else v for v in hoisted]

def vars(args: Tuple[ir.Value, ...]) -> list[ir.Var]:
    """ Filter this list of values, keeping only Vars. """
    return cast(list[ir.Var], list(filter(lambda v: isinstance(v, ir.Var), args)))

def output_vars(aliases: FrozenOrderedSet[Tuple[str, ir.Var]]) -> list[ir.Var]:
    return [alias[1] for alias in aliases]

#--------------------------------------------------
# Visitors/Collectors
#--------------------------------------------------

def collect_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect all Vars starting at this node. """
    return cast(OrderedSet[ir.Var],
        visitor.collect_by_type(ir.Var, *nodes)
    )

def collect_quantified_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect all Vars that are children of Exists and ForAll. """
    return cast(OrderedSet[ir.Var],
        visitor.collect(
            lambda n, parent: isinstance(n, ir.Var) and isinstance(parent, (ir.Exists, ir.ForAll)),
            *nodes)
    )

def collect_aggregate_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect vars that are declared by aggregates in Rel (projection + over). """
    return cast(OrderedSet[ir.Var],
        # TODO - when dealing with multiple aggregations we will need to consider groupbys
        visitor.collect(
            lambda n, parent:
                # parent of var is an aggregate and either
                isinstance(parent, ir.Aggregate) and isinstance(n, ir.Var) and (
                # var is in the projection
                n in parent.projection or
                # var is an input to the aggregation's relation
                is_aggregate_input(n, parent)
                ),
            *nodes)
    )

def collect_implicit_vars(*nodes: ir.Node) -> OrderedSet[ir.Var]:
    """ Collect vars except the quantified and aggregate vars. """
    if not nodes:
        return util.ordered_set()
    return collect_vars(*nodes) - collect_quantified_vars(*nodes) - collect_aggregate_vars(*nodes)


#--------------------------------------------------
# Useful node categories
#--------------------------------------------------

BINDERS = (ir.Lookup, ir.Not, ir.Exists, ir.ForAll)
COMPOSITES = (ir.Logical, ir.Sequence, ir.Union, ir.Match, ir.Until, ir.Wait)
EFFECTS = (ir.Update, ir.Output)



from abc import ABC, abstractmethod
from typing import Union, Sequence

from colorama import Style, Fore

from . import ir

class StyleConfig(ABC):
    @abstractmethod
    def SIND(self, ) -> str: pass

    @abstractmethod
    def LPAREN(self, ) -> str: pass
    @abstractmethod
    def RPAREN(self, ) -> str: pass

    @abstractmethod
    def LBRACKET(self, ) -> str: pass
    @abstractmethod
    def RBRACKET(self, ) -> str: pass

    # String of level indentations for LLQP.
    @abstractmethod
    def indentation(self, level: int) -> str: pass

    # Styled keyword x.
    @abstractmethod
    def kw(self, x: str) -> str: pass

    # Styled user provided name, e.g. variables.
    @abstractmethod
    def uname(self, x: str) -> str: pass

    # Styled type annotation, e.g. ::INT.
    @abstractmethod
    def type_anno(self, x: str) -> str: pass

# Some basic components and how they are to be printed.
class Unstyled(StyleConfig):
    # Single INDentation.
    def SIND(self, ): return "    "

    def LPAREN(self, ): return "("
    def RPAREN(self, ): return ")"

    def LBRACKET(self, ): return "["
    def RBRACKET(self, ): return "]"

    # String of level indentations for LLQP.
    def indentation(self, level: int) -> str:
        return self.SIND() * level

    # Styled keyword x.
    def kw(self, x: str) -> str:
        return x

    # Styled user provided name, e.g. variables.
    def uname(self, x: str) -> str:
        return x

    # Styled type annotation, e.g. ::INT.
    def type_anno(self, x: str) -> str:
        return x

class Styled(StyleConfig):
    def SIND(self, ): return "    "

    def LPAREN(self, ): return Style.DIM + "(" + Style.RESET_ALL
    def RPAREN(self, ): return Style.DIM + ")" + Style.RESET_ALL

    def LBRACKET(self, ): return Style.DIM + "[" + Style.RESET_ALL
    def RBRACKET(self, ): return Style.DIM + "]" + Style.RESET_ALL

    def indentation(self, level: int) -> str:
        return self.SIND() * level

    def kw(self, x: str) -> str:
        return Fore.YELLOW + x + Style.RESET_ALL

    def uname(self, x: str) -> str:
        return Fore.WHITE + x + Style.RESET_ALL

    # Styled type annotation, e.g. ::INT.
    def type_anno(self, x: str) -> str:
        return Style.DIM + x + Style.RESET_ALL

# Call to_llqp on all nodes, each of which with indent_level, separating them
# by delim.
def list_to_llqp(nodes: Sequence[ir.LqpNode], indent_level: int, delim: str, styled: bool) -> str:
    return delim.join(map(lambda n: to_llqp(n, indent_level, styled), nodes))

# Produces "(terms term1 term2 ...)" (all on one line) indented at indent_level.
def terms_to_llqp(terms: Sequence[ir.Term], indent_level: int, styled: bool) -> str:
    conf = Styled() if styled else Unstyled()

    ind = conf.indentation(indent_level)

    llqp = ""
    if len(terms) == 0:
        llqp = ind + conf.LPAREN() + conf.kw("terms") + conf.RPAREN()
    else:
        llqp = ind + conf.LPAREN() + conf.kw("terms") + " " + list_to_llqp(terms, 0, " ", styled) + conf.RPAREN()

    return llqp

def program_to_llqp(node: ir.LqpProgram, styled: bool = True) -> str:
    conf = Styled() if styled else Unstyled()

    # TODO: is this true? and in general for the other things can they be missing?
    reads_portion = ""
    if len(node.outputs) == 0:
        reads_portion += conf.indentation(2) + conf.LPAREN() + conf.kw("reads") + conf.RPAREN() +" ;; no outputs" + "\n"
    else:
        reads_portion += conf.indentation(2) + conf.LPAREN() + conf.kw("reads") + "\n"

        for (name, rel_id) in node.outputs:
            reads_portion +=\
                f"{conf.indentation(3)}" +\
                conf.LPAREN() + f"{conf.uname(name)} " +\
                f"{to_llqp(rel_id, 0, styled)}" +\
                conf.RPAREN()

        reads_portion += conf.RPAREN()

    delim = "\n\n"
    writes_portion = f"{list_to_llqp(node.defs, 5, delim, styled)}"


    return\
    conf.indentation(0) + conf.LPAREN() + conf.kw("transaction") + "\n" +\
    conf.indentation(1) + conf.LPAREN() + conf.kw("epoch") + "\n" +\
    conf.indentation(2) + conf.LPAREN() + conf.kw("local_writes") + "\n" +\
    conf.indentation(3) + conf.LPAREN() + conf.kw("define") + "\n" +\
    conf.indentation(4) + conf.LPAREN() + conf.kw("fragment") + " " + conf.uname(":f1") + "\n" +\
    writes_portion +\
    conf.RPAREN() + conf.RPAREN() + conf.RPAREN() +\
    "\n" +\
    reads_portion +\
    conf.RPAREN() + conf.RPAREN()

def to_llqp(node: Union[ir.LqpNode, ir.PrimitiveType], indent_level: int, styled: bool = True) -> str:
    conf = Styled() if styled else Unstyled()

    ind = conf.indentation(indent_level)
    llqp = ""

    if isinstance(node, ir.Def):
        llqp += ind + conf.LPAREN() + conf.kw("def") + " " + to_llqp(node.name, 0, styled) + "\n"
        llqp += to_llqp(node.body, indent_level + 1, styled) + "\n"
        if len(node.attrs) == 0:
            llqp += ind + conf.SIND() + conf.LPAREN() + conf.kw("attrs") + conf.RPAREN() + conf.RPAREN()
        else:
            llqp += ind + conf.SIND() + conf.LPAREN() + conf.kw("attrs") + "\n"
            llqp += list_to_llqp(node.attrs, indent_level + 2, "\n", styled) + "\n"
            llqp += ind + conf.SIND() + conf.RPAREN() + conf.RPAREN()

    elif isinstance(node, ir.Loop):
        llqp += ind + conf.LPAREN() + conf.kw("loop") + node.temporal_var + "\n"
        llqp += ind + conf.SIND() + conf.LPAREN() + conf.kw("inits") + "\n"
        llqp += list_to_llqp(node.inits, indent_level + 2, "\n", styled) + "\n"
        llqp += ind + conf.SIND() + conf.RPAREN() + "\n"
        llqp += ind + conf.SIND() + conf.LPAREN() + conf.kw("body") + "\n"
        llqp += to_llqp(node.body, indent_level + 2, styled) + "\n"
        llqp += ind + conf.SIND() + conf.RPAREN() + conf.RPAREN()

    elif isinstance(node, ir.Abstraction):
        llqp += ind + conf.LPAREN() + conf.LBRACKET() + list_to_llqp(node.vars, 0, ", ", styled) + conf.RBRACKET() + "\n"
        llqp += to_llqp(node.value, indent_level + 1, styled) + conf.RPAREN()

    elif isinstance(node, ir.Exists):
        llqp += ind + conf.LPAREN() + conf.kw("exists") + " " + conf.LBRACKET() + list_to_llqp(node.vars, 0, ", ", styled) + conf.RBRACKET() + "\n"
        llqp += to_llqp(node.value, indent_level + 1, styled) + conf.RPAREN()

    elif isinstance(node, ir.Reduce):
        llqp += ind + conf.LPAREN() + conf.kw("reduce") + "\n"
        llqp += to_llqp(node.op, indent_level + 1, styled) + "\n"
        llqp += to_llqp(node.body, indent_level + 1, styled) + "\n"
        llqp += terms_to_llqp(node.terms, indent_level + 1, styled) + conf.RPAREN()

    elif isinstance(node, ir.Conjunction):
        llqp += ind + conf.LPAREN() + conf.kw("and") + "\n"
        llqp += list_to_llqp(node.args, indent_level + 1, "\n", styled) + conf.RPAREN()

    elif isinstance(node, ir.Disjunction):
        llqp += ind + conf.LPAREN() + conf.kw("or") + "\n"
        llqp += list_to_llqp(node.args, indent_level + 1, "\n", styled) + conf.RPAREN()

    elif isinstance(node, ir.Not):
        llqp += ind + conf.LPAREN() + conf.kw("not") + "\n"
        llqp += to_llqp(node.arg, indent_level + 1, styled) + conf.RPAREN()

    elif isinstance(node, ir.Ffi):
        llqp += ind + conf.LPAREN() + conf.kw("ffi") + " " + ":" + node.name + "\n"
        llqp += ind + conf.SIND() + conf.LPAREN() + conf.kw("args") + "\n"
        llqp += list_to_llqp(node.args, indent_level + 2, "\n", styled) + "\n"
        llqp += ind + conf.SIND() + conf.RPAREN() + "\n"
        llqp += terms_to_llqp(node.terms, indent_level + 1, styled) + conf.RPAREN()

    elif isinstance(node, ir.Atom):
        llqp += f"{ind}{conf.LPAREN()}{conf.kw('atom')} {to_llqp(node.name, 0, styled)} {list_to_llqp(node.terms, 0, ' ', styled)}{conf.RPAREN()}"

    elif isinstance(node, ir.Pragma):
        llqp += f"{ind}{conf.LPAREN()}{conf.kw('pragma')} :{conf.uname(node.name)} {terms_to_llqp(node.terms, 0, styled)}{conf.RPAREN()}"

    elif isinstance(node, ir.Primitive):
        llqp += f"{ind}{conf.LPAREN()}{conf.kw('primitive')} :{conf.uname(node.name)} {list_to_llqp(node.terms, 0, ' ', styled)}{conf.RPAREN()}"

    elif isinstance(node, ir.JustTrue):
        llqp += ind + conf.LPAREN() + conf.kw("true") + conf.RPAREN()

    elif isinstance(node, ir.JustFalse):
        llqp += ind + conf.LPAREN() + conf.kw("false") + conf.RPAREN()

    elif isinstance(node, ir.RelAtom):
        llqp += f"{ind}{conf.LPAREN()}{conf.kw('relatom')} {to_llqp(node.sig, 0, styled)} {list_to_llqp(node.terms, 0, ' ', styled)}{conf.RPAREN()}"

    elif isinstance(node, ir.Var):
        llqp += ind + conf.uname(node.name) + conf.type_anno("::" + to_llqp(node.type, 0, styled))

    elif isinstance(node, ir.Constant):
        llqp += ind
        if isinstance(node.value, str):
            llqp += "\"" + node.value + "\""
        else:
            # suffices to just dump the value?
            llqp += str(node.value)

    elif isinstance(node, ir.Attribute):
        llqp += ind
        llqp += conf.LPAREN() + conf.kw("attribute") + " "
        llqp += ":" + node.name + " "
        if len(node.args) == 0:
            llqp += conf.LPAREN() + conf.kw("args") + conf.RPAREN()
        else:
            llqp += conf.LPAREN() + conf.kw("args") + " "
            llqp += list_to_llqp(node.args, 0, " ", styled)
            llqp += conf.RPAREN()
        llqp += conf.RPAREN()

    elif isinstance(node, ir.RelationId):
        llqp += f"{ind}:{conf.uname(str(node.id))}"

    elif isinstance(node, ir.RelationSig):
        llqp += ind
        llqp += conf.LPAREN() + conf.kw("sig") + " "
        llqp += ":" + conf.uname(node.name)
        types = ""
        for ty in node.types:
            types += "/" + to_llqp(ty, 0, styled)
        llqp += conf.type_anno(types)
        llqp += conf.RPAREN()

    elif isinstance(node, ir.PrimitiveType):
        llqp += ind + node.name

    else:
        raise NotImplementedError(f"to_llqp not implemented for {type(node)}.")

    return llqp



from .. import dsl
from ..std.graphs import unwrap

# --------------------------------------------------
# Ego network
# --------------------------------------------------
def ego_network(graph, node, hops):
    if not isinstance(hops, int):
        raise TypeError(f"'hops' must be an integer, got {type(hops).__name__} instead")
    elif hops < 0:
        raise ValueError(f"'hops' must be non-negative, got {hops}")

    a, b = dsl.create_vars(2)
    dsl.global_ns.graphlib_experimental.ego_network(graph, unwrap(node), hops, a, b)
    la = graph.compute._lookup(a)
    lb = graph.compute._lookup(b)
    return (la, lb)
